#include "../../alg/alg_base.hpp"
#include "../../util/util_registry.hpp"

#include "alg_canny251_gladjenje.hpp"
#include "alg_canny251_gradijent.hpp"
#include "alg_canny251_stanjivanje.hpp"
#include "alg_canny251_prag.hpp"

#include "../../img/img_vector.hpp"
#include "../../img/access/img_access_copy.hpp"
#include "../../img/img_util.hpp"
#include "../../ipp/ipp_equalize.hpp"
#include "../../cli/cli_wrapAbstract.hpp"

#include <sstream>
#include <time.h>



////////////////////////////////////
// alg_canny251 declaration

class alg_canny251:
  public alg_base
  {
    public:	
      alg_canny251(const CtorArg& arg);	
      virtual ~alg_canny251();
    public:	
      virtual char const* name(){return s_name();}
      static char const* s_name(){return "canny251";}
    public:
      virtual void process(
        const img_vectorAbstract& src, 
        const win_event_vectorAbstract& events,
        int msDayUtc); 
    public:	
      virtual int countDst(); // >=1!
      virtual const img_wrap& imgDst(int id);
    public:	
      virtual void config(cli_wrapAbstract& cli);
      virtual void config(char const* cfgstr);
    private:
      std::string dumpParams();

    private:	
      img_vector imgs_;
      
      img_wrap imgGray;
      img_wrap imgFirstPassSmooth;
      img_wrap imgSmooth;	
      img_wrap imgGrad;
      img_wrap imgPhi;
      img_wrap imgSuppressed;
      img_wrap imgThreshold;

      double sigma_;
      double thHi_;
      double thLo_;
  };

////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<alg_base,alg_canny251> creator;
} // unnamed namespace end


////////////////////////////////////
// alg_canny251 implementation

alg_canny251::alg_canny251(const CtorArg& arg):
  imgs_(8),
  imgGray(arg.fmt),
  imgFirstPassSmooth(arg.fmt),
  imgSmooth(arg.fmt),
  imgGrad(arg.fmt),
  imgPhi(arg.fmt),
  imgSuppressed(arg.fmt),
  imgThreshold(arg.fmt)
  { 
    // sigma - parametar gausove razdiobe
    // thHi - gornja granica kao postotak
    //	thLo - doljnja granica kao postotak
    config("1 0.95 0.70");
  }

alg_canny251::~alg_canny251(){}

void alg_canny251::process(
  const img_vectorAbstract& src,
  const win_event_vectorAbstract&,
  int)
  {
    img_wrap& imgSrc(imgs_[0]);
    imgSrc=src[0];

    //formatiranje slika na velicinu originala (grayscale)
    imgGray.reformat(img_fmt(src[0].fmt(),1));
    imgFirstPassSmooth.reformat(img_fmt(src[0].fmt(),8));
    imgSmooth.reformat(img_fmt(src[0].fmt(),8));
    imgGrad.reformat(img_fmt(src[0].fmt(),8));
    imgPhi.reformat(img_fmt(src[0].fmt(),8));
    imgSuppressed.reformat(img_fmt(src[0].fmt(),8));
    imgThreshold.reformat(img_fmt(src[0].fmt(),8));

    //kopiranje izvorisne slike u imgGray zadanim formatom
    img_access_copy(src[0], imgGray);

    //deklaracija margine koja pazi na "obradeni" rub slike
    int margina=0;

    //timer za pracenje trajanja pojedine faze
    int clo_ = clock(); 

    //u slucaju da je sigma 0, slika se ne zagladuje
    if(sigma_ == 0)
      {
        img_access_copy(imgGray, imgFirstPassSmooth);
        img_access_copy(imgGray, imgSmooth);
      }
    else
      {
        margina = filterGaussSeparatedDouble(sigma_, 
                                             imgGray,
                                             imgFirstPassSmooth,
                                             imgSmooth);
      }
    std::cerr<<"\nTrajanje faza:\n gladjenje:           "<<clock()-clo_<<"ms\n";
    clo_ = clock();

    //pozivi ostalih faza obrade slike
    margina = gradientDouble(margina, 
                             imgSmooth, 
                             imgGrad, 
                             imgPhi);
    std::cerr<<" izracun gradijenata: "<<clock()-clo_<<"ms\n";
    clo_ = clock();

    margina = stanjivanjeDouble(margina,
                                imgGrad,
                                imgPhi,
                                imgSuppressed);
    std::cerr <<" stanjivanje:         " <<clock()-clo_<<"ms\n";
    clo_ = clock();

    //"ciscenje" slike u imgThreshold kako 
    //ne bi doslo do precrtavanja
    img_util::fillBits(imgThreshold, 0);  
    usporedbaSaPragom(thHi_, 
                      thLo_,
                      margina, 
                      imgSuppressed, 
                      imgThreshold);
    std::cerr<<" usporedba s pragom:  "<<clock()-clo_<<"ms\n";
    clo_ = clock();

    //spremanje obradjenih slika u "slike za iscrtavanje"
    ipp::equalize(imgGray, imgs_[1]);
    ipp::equalize(imgFirstPassSmooth, imgs_[2]);
    ipp::equalize(imgSmooth, imgs_[3]);
    ipp::equalize(imgGrad, imgs_[4]);
    ipp::equalize(imgPhi, imgs_[5]);
    ipp::equalize(imgSuppressed, imgs_[6]);
    ipp::equalize(imgThreshold, imgs_[7]);     
    std::cerr<<" iscrtavanje slika:   "<<clock()-clo_<<"ms\n\n";

  }

int alg_canny251::countDst(){
  return static_cast<int>(imgs_.size());
}

const img_wrap& alg_canny251::imgDst(int id){
  checkId(id, "alg_canny251::imgDst");
  return imgs_[id];
}

void alg_canny251::config(cli_wrapAbstract& cli){
  cli.put(dumpParams().c_str());
  char const* buf=cli.get();
  config(buf);
  cli.put(dumpParams().c_str());
}

void alg_canny251::config(const char* cfgstr){
  std::istringstream iss(cfgstr);	
  iss >>sigma_;  //pridruzivanje dijelova config polja
  iss >>thHi_;   //parametrima potrebnim za rad algoritma
  iss >>thLo_;
}

std::string alg_canny251::dumpParams(){
  std::ostringstream os;
  os <<"Current parameters: sigma=" <<sigma_ <<"\n";
  os <<"                    thHi=" <<thHi_ <<"\n";
  os <<"                    thLo=" <<thLo_ <<"\n";
  return os.str();
}
