#include "../../img/io/img_io_load.hpp"
#include "../../img/img_wrap.hpp"
#include "../../img/img_util.hpp"
#include "../../util/util_registry.hpp"

#include <png.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

#include <stdlib.h>
#include <stdio.h>


/******************************************************
*
** img_io_load_png.cpp  by Sinisa Segvic
*
** reading images from png file format
** to a img_wrap structure
*
** depends on:
**   img_wrap, img_io_load, img_util, 
**   util_Endian, <libpng>
*
*******************************************************/


class img_io_load_png:
  public img_io_load
{
public:
  img_io_load_png(const CtorArg& dummy);
  virtual ~img_io_load_png();

  static char const* s_name(){return "png";}
  virtual char const* name(){return s_name();}
  
  virtual void operator()(char const* name, img_wrap&);
};


////////////////////////////////////
// static registry creator object

namespace{
  util::RegistryCreator<img_io_load,img_io_load_png> creator;
} // unnamed namespace end


////////////////////////////////////
// local helpers

namespace{
  struct PngThrower{
    PngThrower(char const* name){
      context_="img_io_load_png (";
      context_+=name;
      context_+=") - ";
    }

    void operator()(const std::string &msg){
      throw std::runtime_error(context_+msg);
    }

    std::string context_;
  };
  

  struct PngFileHolder{
    PngFileHolder(char const* name){
      pf= fopen(name, "rb");
    }
    ~PngFileHolder(){
      if (pf) fclose(pf);
    }

    FILE *pf;
  };
  
  struct PngMemHolder{
    PngMemHolder(){
      png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,0,0,0);
      info_ptr = png_create_info_struct(png_ptr);
    }
    ~PngMemHolder(){
      png_destroy_read_struct(&png_ptr, 0, 0);
    }

    png_structp png_ptr;
    png_infop info_ptr;
  };
}


////////////////////////////////////
// img_io_load_png

img_io_load_png::img_io_load_png(const CtorArg&){}
img_io_load_png::~img_io_load_png(){}




void img_io_load_png::operator()(
  char const* name, 
  img_wrap& img)
{
  PngThrower error(name);


  // Check whether a file is PNG
  PngFileHolder pfh(name);
  if (!pfh.pf) error("file access error");
  const unsigned int PNG_BYTES_TO_CHECK=4;
  png_byte buf[PNG_BYTES_TO_CHECK];
  if (fread(buf, 1, PNG_BYTES_TO_CHECK, pfh.pf) != PNG_BYTES_TO_CHECK)
    error("magic bytes load error");
  if (png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK))
    error("magic bytes check error");


  // Create memory structures  
  PngMemHolder pmh;
  if (!pmh.png_ptr || !pmh.info_ptr) {
    error(" memory error");
  }


  // Read the image
  if (setjmp(png_jmpbuf(pmh.png_ptr))) {
    /* If we get here, we had a problem reading the file */
    error("image load error");
  }
  png_init_io(pmh.png_ptr, pfh.pf);
  png_set_sig_bytes(pmh.png_ptr, PNG_BYTES_TO_CHECK);
  png_set_expand(pmh.png_ptr);
  png_read_png(pmh.png_ptr, pmh.info_ptr, PNG_TRANSFORM_IDENTITY, 0);

  // Check the format
  png_uint_32 width, height;
  int         bit_depth, color_type, interlace_type;
  png_get_IHDR(pmh.png_ptr, pmh.info_ptr, 
    &width, &height, 
    &bit_depth, &color_type,
    &interlace_type, 0, 0);
  //std::cerr <<width<<'x'<<height<<'x'<<bit_depth <<": "
  //          <<color_type <<"," <<interlace_type <<"\n";
  if (bit_depth!=8 || (color_type!=PNG_COLOR_TYPE_GRAY && 
                       color_type!=PNG_COLOR_TYPE_RGB))
  {
    error("unsupported image format");
  }
  int szPixel=color_type==PNG_COLOR_TYPE_GRAY?1:3;
  img.reformat(img_fmt(width,height, szPixel));
    

  // transfer data
  png_bytep *row_pointers = png_get_rows(pmh.png_ptr, pmh.info_ptr);
  for (unsigned int i=0; i<height; ++i){
    img.setLine(height-i-1, row_pointers[i]);
  }

}
